
__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import fractions
import tempfile
import collections

import radical.utils as ru


# 'enum' for launch method types
LM_NAME_APRUN         = 'APRUN'
LM_NAME_CCMRUN        = 'CCMRUN'
LM_NAME_DPLACE        = 'DPLACE'
LM_NAME_FORK          = 'FORK'
LM_NAME_IBRUN         = 'IBRUN'
LM_NAME_MPIEXEC       = 'MPIEXEC'
LM_NAME_MPIRUN        = 'MPIRUN'
LM_NAME_MPIRUN_MPT    = 'MPIRUN_MPT'
LM_NAME_MPIRUN_CCMRUN = 'MPIRUN_CCMRUN'
LM_NAME_MPIRUN_DPLACE = 'MPIRUN_DPLACE'
LM_NAME_MPIRUN_RSH    = 'MPIRUN_RSH'
LM_NAME_ORTE          = 'ORTE'
LM_NAME_ORTE_LIB      = 'ORTE_LIB'
LM_NAME_POE           = 'POE'
LM_NAME_RUNJOB        = 'RUNJOB'
LM_NAME_RSH           = 'RSH'
LM_NAME_SSH           = 'SSH'
LM_NAME_YARN          = 'YARN'
LM_NAME_SPARK         = 'SPARK'


# ==============================================================================
#
class LaunchMethod(object):

    # List of environment variables that designated Launch Methods should export
    # FIXME: we should find out what env vars are changed or added by 
    #        cud.pre_exec, and then should also export those.  That would make
    #        our launch script ore complicated though...
    EXPORT_ENV_VARIABLES = [
        'LD_LIBRARY_PATH',
        'PATH',
        'PYTHONPATH',
    ]

    MPI_FLAVOR_OMPI    = 'OMPI'
    MPI_FLAVOR_HYDRA   = 'HYDRA'
    MPI_FLAVOR_UNKNOWN = 'unknown'

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self.name     = type(self).__name__
        self._cfg     = cfg
        self._session = session
        self._log     = self._session._log
        self._log.debug('create LM: %s', type(self))

        # A per-launch_method list of environment to remove from the CU environment
        self.env_removables = []

        self.launch_command = None
        
        self._configure()

        # TODO: This doesn't make too much sense for LM's that use multiple
        #       commands, perhaps this needs to move to per LM __init__.
        if self.launch_command is None:
            raise RuntimeError("Launch command not found for LaunchMethod '%s'" % self.name)

        self._log.debug('launch_command: %s', self.launch_command)


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Launch Method.
    #
    @classmethod
    def create(cls, name, cfg, session):

        # Make sure that we are the base-class!
        if cls != LaunchMethod:
            raise TypeError("LaunchMethod factory only available to base class!")

        # In case of undefined LM just return None
        if not name:
            return None

        from .aprun          import APRun
        from .ccmrun         import CCMRun
        from .dplace         import DPlace
        from .fork           import Fork
        from .ibrun          import IBRun
        from .mpiexec        import MPIExec
        from .mpirun         import MPIRun
        from .mpirun_mpt     import MPIRun_MPT
        from .mpirun_ccmrun  import MPIRunCCMRun
        from .mpirun_dplace  import MPIRunDPlace
        from .mpirun_rsh     import MPIRunRSH
        from .orte           import ORTE
        from .orte_lib       import ORTELib
        from .poe            import POE
        from .runjob         import Runjob
        from .rsh            import RSH
        from .ssh            import SSH
        from .yarn           import Yarn
        from .spark          import Spark

        try:
            impl = {
                LM_NAME_APRUN         : APRun,
                LM_NAME_CCMRUN        : CCMRun,
                LM_NAME_DPLACE        : DPlace,
                LM_NAME_FORK          : Fork,
                LM_NAME_IBRUN         : IBRun,
                LM_NAME_MPIEXEC       : MPIExec,
                LM_NAME_MPIRUN        : MPIRun,
                LM_NAME_MPIRUN_MPT    : MPIRun_MPT,
                LM_NAME_MPIRUN_CCMRUN : MPIRunCCMRun,
                LM_NAME_MPIRUN_DPLACE : MPIRunDPlace,
                LM_NAME_MPIRUN_RSH    : MPIRunRSH,
                LM_NAME_ORTE          : ORTE,
                LM_NAME_ORTE_LIB      : ORTELib,
                LM_NAME_POE           : POE,
                LM_NAME_RUNJOB        : Runjob,
                LM_NAME_RSH           : RSH,
                LM_NAME_SSH           : SSH,
                LM_NAME_YARN          : Yarn,
                LM_NAME_SPARK         : Spark
            }[name]
            return impl(cfg, session)

        except KeyError:
            session._log.exception("LaunchMethod '%s' unknown or defunct" % name)

        except Exception as e:
            session._log.exception("LaunchMethod cannot be used: %s!" % e)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def lrms_config_hook(cls, name, cfg, lrms, logger, profiler):
        """
        This hook will allow the LRMS to perform launch methods specific
        configuration steps.  The LRMS layer MUST ensure that this hook is
        called exactly once (globally).  This will be a NOOP for LMs which do
        not overload this method.  Exceptions fall through to the LRMS.
        """

        # Make sure that we are the base-class!
        if cls != LaunchMethod:
            raise TypeError("LaunchMethod config hook only available to base class!")

        from .fork           import Fork
        from .orte           import ORTE
        from .yarn           import Yarn
        from .spark          import Spark

        impl = {
            LM_NAME_FORK          : Fork,
            LM_NAME_ORTE          : ORTE,
            LM_NAME_YARN          : Yarn,
            LM_NAME_SPARK         : Spark
        }.get(name)

        if not impl:
            logger.info('no LRMS config hook defined for LaunchMethod %s' % name)
            return None

        logger.info('call LRMS config hook for LaunchMethod %s: %s' % (name, impl))
        return impl.lrms_config_hook(name, cfg, lrms, logger, profiler)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def lrms_shutdown_hook(cls, name, cfg, lrms, lm_info, logger, profiler):
        """
        This hook is symmetric to the config hook above, and is called during
        shutdown sequence, for the sake of freeing allocated resources.
        """

        # Make sure that we are the base-class!
        if cls != LaunchMethod:
            raise TypeError("LaunchMethod shutdown hook only available to base class!")

        from .orte           import ORTE
        from .yarn           import Yarn
        from .spark          import Spark

        impl = {
            LM_NAME_ORTE          : ORTE,
            LM_NAME_YARN          : Yarn,
            LM_NAME_SPARK         : Spark
        }.get(name)

        if not impl:
            logger.info('no LRMS shutdown hook defined for LaunchMethod %s' % name)
            return None

        logger.info('call LRMS shutdown hook for LaunchMethod %s: %s' % (name, impl))
        return impl.lrms_shutdown_hook(name, cfg, lrms, lm_info, logger, profiler)


    # --------------------------------------------------------------------------
    #
    def _configure(self):
        raise NotImplementedError("_configure() not implemented for LaunchMethod: %s." % self.name)


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):
        raise NotImplementedError("construct_command() not implemented for LaunchMethod: %s." % self.name)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def _create_hostfile(cls, all_hosts, separator=' ', impaired=False):

        # Open appropriately named temporary file
        handle, filename = tempfile.mkstemp(prefix='rp_hostfile', dir=os.getcwd())

        if not impaired:
            #
            # Write "hostN x\nhostM y\n" entries
            #

            # Create a {'host1': x, 'host2': y} dict
            counter = collections.Counter(all_hosts)
            # Convert it into an ordered dict,
            # which hopefully resembles the original ordering
            count_dict = collections.OrderedDict(sorted(counter.items(), key=lambda t: t[0]))

            for (host, count) in count_dict.iteritems():
                os.write(handle, '%s%s%d\n' % (host, separator, count))

        else:
            #
            # Write "hostN\nhostM\n" entries
            #
            for host in all_hosts:
                os.write(handle, '%s\n' % host)

        # No longer need to write
        os.close(handle)

        # Return the filename, caller is responsible for cleaning up
        return filename


    # --------------------------------------------------------------------------
    #
    @classmethod
    def _compress_hostlist(cls, all_hosts):

        # Return gcd of a list of numbers
        def gcd_list(l):
            return reduce(fractions.gcd, l)

        # Create a {'host1': x, 'host2': y} dict
        count_dict = dict(collections.Counter(all_hosts))
        # Find the gcd of the host counts
        host_gcd = gcd_list(set(count_dict.values()))

        # Divide the host counts by the gcd
        for host in count_dict:
            count_dict[host] /= host_gcd

        # Recreate a list of hosts based on the normalized dict
        hosts = []
        [hosts.extend([host] * count)
                for (host, count) in count_dict.iteritems()]
        # Esthetically sort the list, as we lost ordering by moving to a dict/set
        hosts.sort()

        return hosts


    # --------------------------------------------------------------------------
    #
    def _create_arg_string(self, args):

        # unit Arguments (if any)
        arg_string = ''
        if args:
            for arg in args:
                if not arg:
                    # ignore empty args
                    continue

                if arg in ['>', '>>', '<', '<<', '|', '||', '&&', '&']:
                    # Don't quote shell direction arguments, etc.
                    arg_string += '%s ' % arg
                    continue

                if any([c in arg for c in ['?', '*']]):
                    # Don't quote arguments with wildcards
                    arg_string += '%s ' % arg
                    continue

                arg = arg.replace('"', '\\"')    # Escape all double quotes
                if arg[0] == arg[-1] == "'" :    # If a string is between outer single quotes,
                    arg_string += '%s ' % arg    # ... pass it as is.
                else:
                    arg_string += '"%s" ' % arg  # Otherwise return between double quotes.

        return arg_string


    # --------------------------------------------------------------------------
    #
    def _get_mpi_info(self, exe):
        '''
        returns version and flavor of MPI version.
        '''

        version = None
        flavor  = self.MPI_FLAVOR_UNKNOWN

        out, err, ret = ru.sh_callout('%s -v' % exe)

        if ret:
            out, err, ret = ru.sh_callout('%s --version' % exe)

        if ret:
            out, err, ret = ru.sh_callout('%s -info' % exe)

        if not ret:
            for line in out.splitlines():
                if 'hydra build details:' in line.lower():
                    version = line.split(':', 1)[1].strip()
                    flavor  = self.MPI_FLAVOR_HYDRA
                    break

                if 'mvapich2' in line.lower():
                    version = line
                    flavor  = self.MPI_FLAVOR_HYDRA
                    break

                if 'version:' in line.lower():
                    version = line.split(':', 1)[1].strip()
                    flavor  = self.MPI_FLAVOR_OMPI
                    break

                if '(open mpi):' in line.lower():
                    version = line.split(')', 1)[1].strip()
                    flavor  = self.MPI_FLAVOR_OMPI
                    break

        if not flavor:
            raise RuntimeError('cannot identify MPI flavor [%s]' % exe)

        self._log.debug('mpi version: %s [%s]', version, flavor)

        return version, flavor




# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2017, http://radical.rutgers.edu"
__license__   = "MIT"


import time
from ... import constants as rpc
from .base import AgentSchedulingComponent


# ==============================================================================
#
# Scattered core agent scheduler.
# Finds available cores within the job's allocation.
#
# TODO: For non-MPI multicore tasks, this provides no guarantee that they are
#       on the same host.
#
class Scattered(AgentSchedulingComponent):
    

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self.nodes = None
        AgentSchedulingComponent.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        # TODO: use real core/gpu numbers for non-exclusive reservations

        self.nodes = []
        for node in self._lrms_node_list:
            self.nodes.append({
                'name' : '%s:0' % node,
                'cores': rpc.FREE * self._lrms_cores_per_node,
                'gpus' : rpc.FREE * self._lrms_gpus_per_node
            })


    # --------------------------------------------------------------------------
    #
    # Provide visual presentation of slot status
    #
    def slot_status(self):
        """Returns a multi-line string corresponding to slot status.
        """

        ret = "|"
        for node in self.nodes:
            for core in node['cores']:
                if core == rpc.FREE  : ret += '-'
                else                 : ret += '#'
            ret += ':'
            for gpu in node['gpus']  :
                if gpu == rpc.FREE   : ret += '-'
                else                 : ret += '#'
            ret += '|'

        return ret


    # --------------------------------------------------------------------------
    #
    # Find cores and allocate them if available
    #
    def _allocate_slot(self, cud):

        cores_requested = cud['cpu_processes'] * cud['cpu_threads']
        gpus_requested  = cud['gpus']

        offsets = []
        for offset, node in enumerate(self.nodes):

            if node['state'] == rpc.FREE:
                offsets.append(offset)

            elif len(offsets) == cores_requested:
                break

        if len(offsets) < cores_requested:
            return {}   # allocation failed

        task_slots = [self.nodes[x]['name'] for x in offsets]

        self._change_slot_states(offsets, rpc.BUSY)

        result = {
            'task_slots'   : task_slots,
            'task_offsets' : offsets,
            'lm_info'      : self._lrms_lm_info
        }

        self._log.debug('Slot allocated: %s', result)

        return result


    # --------------------------------------------------------------------------
    #
    # Release cores associated to this slot
    #
    def _release_slot(self, slots):

        if not 'task_offsets' in slots:
            raise RuntimeError('insufficient information to release slots via %s: %s' \
                    % (self.name, slots))

        self._change_slot_states(slots['task_offsets'], rpc.FREE)


    # --------------------------------------------------------------------------
    #
    # Change the reserved state of slots at the offsets(rpc.FREE or rpc.BUSY)
    #
    def _change_slot_states(self, offsets, new_state):

        for offset in offsets:
            self.nodes[offset]['state'] = new_state

# ------------------------------------------------------------------------------




__author__    = "George Chantzialexiou"
__copyright__ = "Copyright 2014, The RADICAL Group"
__license__   = "MIT"

import sys
import random

#------------------------------------------------------------------------------

if __name__ == "__main__":

    args = sys.argv[1:]
    if len(args) < 1:
        print "Usage: python %s needs the  number of the dataset. Please run again" % __file__
        print "python create_dataset.py n"
        sys.exit(-1)
    
    n = int(sys.argv[1])  # k is the number of clusters i want to create
    # read the dataset from dataset.data f
    
    #open the file where we write the results
    data_file = open('dataset2.data', 'w')
    
    
    for i in range(0,n-1):
        a = random.uniform(0, 10000)
        data_file.write(str(a))
        data_file.write(',')
    
    # writing the last element separately, because we don't need to add a coma
    a = random.uniform(0, 10000)
    data_file.write(str(a))
    
    data_file.close()

    sys.exit(0)


#------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import copy
import time
import errno
import Queue
import tempfile
import threading
import traceback

from orte_cffi import ffi, lib as orte_lib

from ....  import pilot as rp
from ...  import states    as rps
from ...  import constants as rpc
from .base import AgentExecutingComponent

# ------------------------------------------------------------------------------
#
def rec_makedir(target):

    # recursive makedir which ignores errors if dir already exists

    try:
        os.makedirs(target)

    except OSError as e:
        # ignore failure on existing directory
        if e.errno == errno.EEXIST and os.path.isdir(os.path.dirname(target)):
            pass
        else:
            raise


# ==============================================================================
#
@ffi.def_extern()
def launch_cb(task, jdata, status, cbdata):
    return ffi.from_handle(cbdata).unit_spawned_cb(task, status)


# ==============================================================================
#
@ffi.def_extern()
def finish_cb(task, jdata, status, cbdata):
    return ffi.from_handle(cbdata).unit_completed_cb(task, status)


# ==============================================================================
#
class ORTE(AgentExecutingComponent):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        AgentExecutingComponent.__init__(self, cfg, session)

        self._watcher   = None
        self._terminate = threading.Event()


    # --------------------------------------------------------------------------
    #
    def initialize_child(self):

        self._pwd = os.getcwd()

        self.register_input(rps.AGENT_EXECUTING_PENDING,
                            rpc.AGENT_EXECUTING_QUEUE, self.work)

        self.register_output(rps.AGENT_STAGING_OUTPUT_PENDING,
                             rpc.AGENT_STAGING_OUTPUT_QUEUE)

        self.register_publisher (rpc.AGENT_UNSCHEDULE_PUBSUB)
        self.register_subscriber(rpc.CONTROL_PUBSUB, self.command_cb)

        self._cancel_lock    = threading.RLock()
        self._cus_to_cancel  = list()
        self._watch_queue    = Queue.Queue ()

        self._pilot_id = self._cfg['pilot_id']

        self.task_map = {}
        self.task_map_lock = threading.Lock()

        # we needs the LaunchMethods to construct commands.
        assert(self._cfg['task_launch_method'] == \
               self._cfg['mpi_launch_method' ] == \
               "ORTE_LIB"), "ORTE_LIB spawner only works with ORTE_LIB LM's."

        self._task_launcher = rp.agent.LM.create(name    = "ORTE_LIB",
                                                 cfg     = self._cfg,
                                                 session = self._session)
        self._orte_initialized = False
        self._cu_environment   = self._populate_cu_environment()

        self.gtod   = "%s/gtod" % self._pwd
        self.tmpdir = tempfile.gettempdir()


    # --------------------------------------------------------------------------
    #
    def command_cb(self, topic, msg):

        self._log.info('command_cb [%s]: %s', topic, msg)

        cmd = msg['cmd']
        arg = msg['arg']

        if cmd == 'cancel_units':

            self._log.info("cancel_units command (%s)" % arg)
            with self._cancel_lock:
                self._cus_to_cancel.extend(arg['uids'])

        return True


    # --------------------------------------------------------------------------
    #
    def _populate_cu_environment(self):
        """Derive the environment for the cu's from our own environment."""

        # Get the environment of the agent
        new_env = copy.deepcopy(os.environ)

        #
        # Mimic what virtualenv's "deactivate" would do
        #
        old_path = new_env.pop('_OLD_VIRTUAL_PATH', None)
        if old_path:
            new_env['PATH'] = old_path

        # TODO: verify this snippet from:
        # https://github.com/radical-cybertools/radical.pilot/pull/973/files
        # old_ppath = new_env.pop('_OLD_VIRTUAL_PYTHONPATH', None)
        # if old_ppath:
        #     new_env['PYTHONPATH'] = old_ppath

        old_home = new_env.pop('_OLD_VIRTUAL_PYTHONHOME', None)
        if old_home:
            new_env['PYTHON_HOME'] = old_home

        old_ps = new_env.pop('_OLD_VIRTUAL_PS1', None)
        if old_ps:
            new_env['PS1'] = old_ps

        new_env.pop('VIRTUAL_ENV', None)

        # Remove the configured set of environment variables from the
        # environment that we pass to Popen.
        for e in new_env.keys():
            env_removables = list()
            if self._task_launcher:
                env_removables += self._task_launcher.env_removables
            for r in  env_removables:
                if e.startswith(r):
                    new_env.pop(e, None)

        return new_env

    # --------------------------------------------------------------------------
    #
    def work(self, units):

        if not isinstance(units, list):
            units = [units]

        self.advance(units, rps.AGENT_EXECUTING, publish=True, push=False)

        for unit in units:
            self._handle_unit(unit)


    # --------------------------------------------------------------------------
    #
    def _handle_unit(self, cu):

        # prep stdout/err so that we can append w/o checking for None
        cu['stdout'] = ''
        cu['stderr'] = ''

        if not self._orte_initialized:
            self._log.debug("ORTE not yet initialized!")
            ret = self.init_orte(cu)
            if ret != 0:
                self._log.debug("ORTE initialisation failed!")
            else:
                self._log.debug("ORTE initialisation succeeded!")

        try:
            launcher = self._task_launcher

            if not launcher:
                raise RuntimeError("no launcher (mpi=%s)" % cu['description']['mpi'])

            self._log.debug("Launching unit with %s (%s).", 
                            launcher.name, launcher.launch_command)

            assert(cu['slots']), 'unit unscheduled'
            # Start a new subprocess to launch the unit
            self.spawn(launcher=launcher, cu=cu)

        except Exception as e:
            # append the startup error to the units stderr.  This is
            # not completely correct (as this text is not produced
            # by the unit), but it seems the most intuitive way to
            # communicate that error to the application/user.
            self._log.exception("error running CU: %s", str(e))
            cu['stderr'] += "\nPilot cannot start compute unit:\n%s\n%s" \
                            % (str(e), traceback.format_exc())

            # Free the Slots, Flee the Flots, Ree the Frots!
            if cu['slots']:
                self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, cu)

            self.advance(cu, rps.FAILED, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def unit_spawned_cb(self, task, status):

        with self.task_map_lock:
            cu = self.task_map[task]
        uid = cu['uid']

        if status:
            with self.task_map_lock:
                del self.task_map[task]

            # unit launch failed
            self._prof.prof('exec_fail', uid=uid)
            self._log.error("unit %s startup failed: %s", uid, status)
            self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, cu)

            cu['target_state'] = rps.FAILED
            self.advance(cu, rps.AGENT_STAGING_OUTPUT_PENDING, 
                         publish=True, push=True)

        else:
            cu['started'] = time.time()
            self._log.debug("unit %s startup ok", uid)
            self.advance(cu, rps.AGENT_EXECUTING, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def unit_completed_cb(self, task, exit_code):

        timestamp = time.time()

        with self.task_map_lock:
            cu = self.task_map[task]
            del self.task_map[task]

        self._prof.prof('exec_stop', uid=cu['uid'])
        self._log.info("unit %s finished [%s]", cu['uid'], exit_code)

        cu['exit_code'] = exit_code
        cu['finished']  = timestamp

        self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, cu)

        if exit_code != 0:
            # unit failed - fail after staging output
            cu['target_state'] = rps.FAILED

        else:
            # unit finished cleanly.  We always move to stageout, even if there 
            # are no staging directives -- at the very least, we'll upload
            # stdout/stderr
            cu['target_state'] = rps.DONE

        self.advance(cu, rps.AGENT_STAGING_OUTPUT_PENDING, 
                     publish=True, push=True)


    # --------------------------------------------------------------------------
    #
    def init_orte(self, cu):

        # FIXME: it feels as a hack to get the DVM URI from the CU

        slots = cu['slots']

        if 'lm_info' not in slots:
            raise RuntimeError('No lm_info to init via %s: %s' \
                               % (self.name, slots))

        if not slots['lm_info']:
            raise RuntimeError('lm_info missing for %s: %s' \
                               % (self.name, slots))

        if 'dvm_uri' not in slots['lm_info']:
            raise RuntimeError('dvm_uri not in lm_info for %s: %s' \
                               % (self.name, slots))

        dvm_uri = slots['lm_info']['dvm_uri']

        # Notify orte that we are using threads and that we require mutexes
        orte_lib.opal_set_using_threads(True)

        argv_keepalive = [
            ffi.new("char[]", "RADICAL-Pilot"), # will be stripped off by lib
            ffi.new("char[]", "--hnp"), ffi.new("char[]", str(dvm_uri)),
            ffi.NULL, # required
        ]
        argv = ffi.new("char *[]", argv_keepalive)
        ret = orte_lib.orte_submit_init(3, argv, ffi.NULL)

        self._myhandle = ffi.new_handle(self)
        self._orte_initialized = True

        return ret

    # --------------------------------------------------------------------------
    #
    def spawn(self, launcher, cu):

        # NOTE: see documentation of cu['sandbox'] semantics in the ComputeUnit
        #       class definition.
        sandbox = '%s/%s' % (self._pwd, cu['uid'])

        if False:
            cu_tmpdir = '%s/%s' % (self.tmpdir, cu['uid'])
        else:
            cu_tmpdir = sandbox

        rec_makedir(cu_tmpdir)

        # TODO: pre_exec
        # # Before the Big Bang there was nothing
        # if cu['description']['pre_exec']:
        #     fail = ' (echo "pre_exec failed"; false) || exit'
        #     pre  = ''
        #     for elem in cu['description']['pre_exec']:
        #         pre += "%s || %s\n" % (elem, fail)
        #     # Note: extra spaces below are for visual alignment
        #     launch_script.write("# Pre-exec commands\n")
        #     if 'RADICAL_PILOT_PROFILE' in os.environ:
        #         launch_script.write("echo cu_pre_start `%s` >> %s/%s.prof\n"\
        #                           % (cu['gtod'], cu_tmpdir, cu['uid']))
        #     launch_script.write(pre)
        #     if 'RADICAL_PILOT_PROFILE' in os.environ:
        #         launch_script.write("echo cu_pre_stop `%s` >> %s/%s.prof\n" \
        #                           % (cu['gtod'], cu_tmpdir, cu['uid']))

        # TODO: post_exec
        # # After the universe dies the infrared death, there will be nothing
        # if cu['description']['post_exec']:
        #     fail = ' (echo "post_exec failed"; false) || exit'
        #     post = ''
        #     for elem in cu['description']['post_exec']:
        #         post += "%s || %s\n" % (elem, fail)
        #     launch_script.write("# Post-exec commands\n")
        #     if 'RADICAL_PILOT_PROFILE' in os.environ:
        #         launch_script.write("echo cu_post_start `%s` >> %s/%s.prof\n" \
        #                           % (cu['gtod'], cu_tmpdir, cu['uid']))
        #     launch_script.write('%s\n' % post)
        #     if 'RADICAL_PILOT_PROFILE' in os.environ:
        #         launch_script.write("echo cu_post_stop  `%s` >> %s/%s.prof\n" \
        #                           % (cu['gtod'], cu_tmpdir, cu['uid']))


        # The actual command line, constructed per launch-method
        try:
            orte_command, task_command = launcher.construct_command(cu, None)
        except Exception as e:
            msg = "Error in spawner (%s)" % e
            self._log.exception(msg)
            raise RuntimeError(msg)

        # Construct arguments to submit_job
        arg_list = []

        # Take the orte specific commands and split them
        for arg in orte_command.split():
            arg_list.append(ffi.new("char[]", str(arg)))

        # Set the working directory
        arg_list.append(ffi.new("char[]", "--wdir"))
        arg_list.append(ffi.new("char[]", str(cu_tmpdir)))

        # Set RP environment variables
        rp_envs = [
            "RP_SESSION_ID=%s" % self._cfg['session_id'],
            "RP_PILOT_ID=%s" % self._cfg['pilot_id'],
            "RP_AGENT_ID=%s" % self._cfg['agent_name'],
            "RP_SPAWNER_ID=%s" % self.uid,
            "RP_UNIT_ID=%s" % cu['uid']
        ]
        for env in rp_envs:
            arg_list.append(ffi.new("char[]", "-x"))
            arg_list.append(ffi.new("char[]", str(env)))

        # Set pre-populated environment variables
        if self._cu_environment:
            for key,val in self._cu_environment.iteritems():
                arg_list.append(ffi.new("char[]", "-x"))
                arg_list.append(ffi.new("char[]", "%s=%s" % (key, val)))

        # Set environment variables specified for this CU
        if cu['description']['environment']:
            for key,val in cu['description']['environment'].iteritems():
                arg_list.append(ffi.new("char[]", "-x"))
                arg_list.append(ffi.new("char[]", "%s=%s" % (key, val)))

        # Let the orted write stdout and stderr to rank-based output files
        arg_list.append(ffi.new("char[]", "--output-filename"))
        arg_list.append(ffi.new("char[]", "%s:nojobid,nocopy" % str(cu_tmpdir)))

        # Save retval of actual CU application (in case we have post-exec)
        task_command += "; RETVAL=$?"

        # Wrap in (sub)shell for output redirection
        arg_list.append(ffi.new("char[]", "sh"))
        arg_list.append(ffi.new("char[]", "-c"))
        if 'RADICAL_PILOT_PROFILE' in os.environ:
            task_command = "echo script cu_start `%s` >> %s/%s.prof; " \
                         % (self.gtod, cu_tmpdir, cu['uid']) \
                         + "echo script cu_cd_done `%s` >> %s/%s.prof; " \
                         % (self.gtod, cu_tmpdir, cu['uid']) \
                         + "echo script cu_exec_start `%s` >> %s/%s.prof; " \
                         % (self.gtod, cu_tmpdir, cu['uid']) \
                         + task_command \
                         + "; echo script cu_exec_stop `%s` >> %s/%s.prof" \
                         % (self.gtod, cu_tmpdir, cu['uid'])
        arg_list.append(ffi.new("char[]", str("%s; exit $RETVAL" \
                                            % str(task_command))))

        self._log.debug("Launching unit %s via %s %s", cu['uid'], 
                        orte_command, task_command)

        # NULL termination, required by ORTE
        arg_list.append(ffi.NULL)
        argv = ffi.new("char *[]", arg_list)

        # stdout/stderr filenames can't be set with orte
        # TODO: assert here or earlier?
        # assert cu['description'].get('stdout') == None
        # assert cu['description'].get('stderr') == None

        # prepare stdout/stderr
        # TODO: when mpi==True && cores>1 there will be multiple files that need
        #       to be concatenated.
        cu['stdout_file'] = os.path.join(cu_tmpdir, 'rank.0/stdout')
        cu['stderr_file'] = os.path.join(cu_tmpdir, 'rank.0/stderr')

        # Submit to the DVM!
        index = ffi.new("int *")
        with self.task_map_lock:

            self._prof.prof('exec_start', uid=cu['uid'])
            rc = orte_lib.orte_submit_job(argv, index, orte_lib.launch_cb, 
                                          self._myhandle, orte_lib.finish_cb, 
                                          self._myhandle)
            if rc:
                raise Exception("submit job failed with error: %d" % rc)

            self.task_map[index[0]] = cu      # map ORTE index to CU
            self._prof.prof('exec_ok', uid=cu['uid'])

        self._log.debug("Task %d submitted!", cu['uid'])


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__author__    = 'RADICAL Team'
__email__     = 'radical@rutgers.edu'
__copyright__ = 'Copyright 2013-16, RADICAL Research, Rutgers University'
__license__   = 'MIT'


""" Setup script, only usable via pip. """

import re
import os
import sys
import glob
import shutil
import subprocess as sp

name     = 'radical.pilot'
mod_root = 'src/radical/pilot/'

try:
    from setuptools import setup, Command, find_packages
except ImportError as e:
    print("%s needs setuptools to install" % name)
    sys.exit(1)


# ------------------------------------------------------------------------------
#
# versioning mechanism:
#
#   - version:          1.2.3            - is used for installation
#   - version_detail:  v1.2.3-9-g0684b06 - is used for debugging
#   - version is read from VERSION file in src_root, which then is copied to
#     module dir, and is getting installed from there.
#   - version_detail is derived from the git tag, and only available when
#     installed from git.  That is stored in mod_root/VERSION in the install
#     tree.
#   - The VERSION file is used to provide the runtime version information.
#
def get_version(mod_root):
    """
    mod_root
        a VERSION file containes the version strings is created in mod_root,
        during installation.  That file is used at runtime to get the version
        information.
        """

    try:

        version_base   = None
        version_detail = None

        # get version from './VERSION'
        src_root = os.path.dirname(__file__)
        if  not src_root:
            src_root = '.'

        with open(src_root + '/VERSION', 'r') as f:
            version_base = f.readline().strip()

        # attempt to get version detail information from git
        # We only do that though if we are in a repo root dir,
        # ie. if 'git rev-parse --show-prefix' returns an empty string --
        # otherwise we get confused if the ve lives beneath another repository,
        # and the pip version used uses an install tmp dir in the ve space
        # instead of /tmp (which seems to happen with some pip/setuptools
        # versions).
        p = sp.Popen('cd %s ; '
                     'test -z `git rev-parse --show-prefix` || exit -1; '
                     'tag=`git describe --tags --always` 2>/dev/null ; '
                     'branch=`git branch | grep -e "^*" | cut -f 2- -d " "` 2>/dev/null ; '
                     'echo $tag@$branch' % src_root,
                     stdout=sp.PIPE, stderr=sp.STDOUT, shell=True)
        version_detail = str(p.communicate()[0].strip())
        version_detail = version_detail.replace('detached from ', 'detached-')

        # remove all non-alphanumeric (and then some) chars
        version_detail = re.sub('[/ ]+', '-', version_detail)
        version_detail = re.sub('[^a-zA-Z0-9_+@.-]+', '', version_detail)

        if  p.returncode   !=  0  or \
            version_detail == '@' or \
            'not-a-git-repo' in version_detail or \
            'not-found'      in version_detail or \
            'fatal'          in version_detail :
            version = version_base
        elif '@' not in version_base:
            version = '%s-%s' % (version_base, version_detail)
        else:
            version = version_base

        # make sure the version files exist for the runtime version inspection
        path = '%s/%s' % (src_root, mod_root)
        with open(path + "/VERSION", "w") as f:
            f.write(version + "\n")

        sdist_name = "%s-%s.tar.gz" % (name, version)
        sdist_name = sdist_name.replace('/', '-')
        sdist_name = sdist_name.replace('@', '-')
        sdist_name = sdist_name.replace('#', '-')
        sdist_name = sdist_name.replace('_', '-')

        if '--record'    in sys.argv or \
           'bdist_egg'   in sys.argv or \
           'bdist_wheel' in sys.argv    :
          # pip install stage 2 or easy_install stage 1
          #
          # pip install will untar the sdist in a tmp tree.  In that tmp
          # tree, we won't be able to derive git version tags -- so we pack the
          # formerly derived version as ./VERSION
            shutil.move("VERSION", "VERSION.bak")            # backup version
            shutil.copy("%s/VERSION" % path, "VERSION")      # use full version instead
            os.system  ("python setup.py sdist")             # build sdist
            shutil.copy('dist/%s' % sdist_name,
                        '%s/%s'   % (mod_root, sdist_name))  # copy into tree
            shutil.move("VERSION.bak", "VERSION")            # restore version

        with open(path + "/SDIST", "w") as f: 
            f.write(sdist_name + "\n")

        return version_base, version_detail, sdist_name

    except Exception as e :
        raise RuntimeError('Could not extract/set version: %s' % e)


# ------------------------------------------------------------------------------
# check python version. we need >= 2.7, <3.x
if  sys.hexversion < 0x02070000 or sys.hexversion >= 0x03000000:
    raise RuntimeError("%s requires Python 2.x (2.7 or higher)" % name)


# ------------------------------------------------------------------------------
# get version info -- this will create VERSION and srcroot/VERSION
version, version_detail, sdist_name = get_version(mod_root)


# ------------------------------------------------------------------------------
class our_test(Command):
    user_options = []
    def initialize_options(self): pass
    def finalize_options  (self): pass
    def run(self):
        testdir = "%s/tests/" % os.path.dirname(os.path.realpath(__file__))
        retval  = sp.call([sys.executable,
                          '%s/run_tests.py'          % testdir,
                          '%s/configs/basetests.cfg' % testdir])
        raise SystemExit(retval)


# ------------------------------------------------------------------------------
#
def read(fname):
    try :
        return open(fname).read()
    except Exception :
        return ''


df = list()
df.append(('share/%s'                       % name, ['docs/source/events.md']))
df.append(('share/%s/examples'              % name, glob.glob('examples/[01]*.py')))
df.append(('share/%s/examples'              % name, glob.glob('examples/hello*')))
df.append(('share/%s/examples'              % name, glob.glob('examples/*.json')))
df.append(('share/%s/examples/docs'         % name, glob.glob('examples/docs/*')))
df.append(('share/%s/examples/misc'         % name, glob.glob('examples/misc/*')))
df.append(('share/%s/examples/kmeans'       % name, glob.glob('examples/kmeans/*')))
df.append(('share/%s/examples/mandelbrot'   % name, glob.glob('examples/mandelbrot/*')))
df.append(('share/%s/examples/data_staging' % name, glob.glob('examples/data_staging/*')))


# -------------------------------------------------------------------------------
setup_args = {
    'name'               : name,
    'version'            : version,
    'description'        : 'The RADICAL pilot job framework '
                           '(http://radical.rutgers.edu/)',
    'long_description'   : (read('README.md') + '\n\n' + read('CHANGES.md')),
    'author'             : 'RADICAL Group at Rutgers University',
    'author_email'       : 'radical@rutgers.edu',
    'maintainer'         : 'The RADICAL Group',
    'maintainer_email'   : 'radical@rutgers.edu',
    'url'                : 'https://www.github.com/radical-cybertools/radical.utils/',
    'license'            : 'MIT',
    'keywords'           : 'radical pilot job saga',
    'classifiers'        : [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Utilities',
        'Topic :: System :: Distributed Computing',
        'Topic :: Scientific/Engineering',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: Unix'
    ],
    'namespace_packages' : ['radical'],
    'packages'           : find_packages('src'),
    'package_dir'        : {'': 'src'},
    'scripts'            : [
                            'bin/radical-pilot-bson2json',
                            'bin/radical-pilot-cleanup',
                            'bin/radical-pilot-close-session',
                            'bin/radical-pilot-create-static-ve',
                            'bin/radical-pilot-deploy-ompi.sh',
                            'bin/radical-pilot-fetch-db',
                            'bin/radical-pilot-fetch-logfiles',
                            'bin/radical-pilot-fetch-profiles',
                            'bin/radical-pilot-fetch-logfiles',
                            'bin/radical-pilot-fetch-json',
                            'bin/radical-pilot-inspect',
                            'bin/radical-pilot-run-session',
                            'bin/radical-pilot-stats',
                            'bin/radical-pilot-stats.plot',
                            'bin/radical-pilot-version',
                            'bin/radical-pilot-agent',
                            'bin/radical-pilot-agent-statepush'
                           ],
    'package_data'       : {'': ['*.txt', '*.sh', '*.json', '*.gz',
                                 'VERSION', 'SDIST', sdist_name]},
    'cmdclass'           : {
        'test'           : our_test,
                           },
    'install_requires'   : ['saga-python>=0.44',
                            'radical.utils>=0.44',
                            'pymongo',
                            'python-hostlist',
                            'netifaces==0.10.4',
                            'setproctitle',
                            'ntplib',
                            'msgpack-python',
                            'pyzmq'], 
    'extras_require'     : {'autopilot' : ['github3.py']},
    'tests_require'      : ['mock==2.0.0', 'pytest'],
    'test_suite'         : '%s.tests' % name,
    'zip_safe'           : False,
  # 'build_sphinx'       : {
  #     'source-dir'     : 'docs/',
  #     'build-dir'      : 'docs/build',
  #     'all_files'      : 1,
  # },
  # 'upload_sphinx'      : {
  #     'upload-dir'     : 'docs/build/html',
  # },
    # This copies the contents of the examples/ dir under
    # sys.prefix/share/$name
    # It needs the MANIFEST.in entries to work.
  # 'data_files'         : makeDataFiles('share/%s/examples/' % name, 'examples'),
    'data_files'         : df,
}

# ------------------------------------------------------------------------------

setup(**setup_args)

os.system('rm -rf src/%s.egg-info' % name)

# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import radical.utils as ru

from .base import LaunchMethod


# ==============================================================================
#
class DPlace(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        LaunchMethod.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        # dplace: job launcher for SGI systems (e.g. on Blacklight)
        self.launch_command = ru.which('dplace')


    # --------------------------------------------------------------------------
    #
    def construct_command(self, cu, launch_script_hop):

        slots        = cu['slots']
        cud          = cu['description']
        task_exec    = cud['executable']
        task_cores   = cud['cpu_processes']  # FIXME: cpu_threads
        task_args    = cud.get('arguments') or []
        task_argstr  = self._create_arg_string(task_args)

        if 'task_offsets' not in slots :
            raise RuntimeError('insufficient information to launch via %s: %s' \
                    % (self.name, slots))

        task_offsets = slots['task_offsets']        # This needs to revisited 
        assert(len(task_offsets) == 1)              # since slots structure has 
        dplace_offset = task_offsets[0]             # changed

        if task_argstr:
            task_command = "%s %s" % (task_exec, task_argstr)
        else:
            task_command = task_exec


        dplace_command = "%s -c %d-%d %s" % (
            self.launch_command, dplace_offset,
            dplace_offset+task_cores-1, task_command)

        return dplace_command, None


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

from .agent_0        import Agent_0
from .agent_n        import Agent_n

from .rm             import RM
from .lm             import LaunchMethod as LM
from .scheduler      import Scheduler
from .executing      import Executing
from .staging_input  import Input
from .staging_output import Output






__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"

import os
import pprint
import threading

import radical.utils as ru

from ... import utils     as rpu
from ... import states    as rps
from ... import constants as rpc

from .base import UMGRSchedulingComponent, ROLE, ADDED


# the high water mark determines the percentage of unit oversubscription for the
# pilots, in terms of numbers of cores
_HWM = int(os.environ.get('RADICAL_PILOT_BACKFILLING_HWM', 200))

# we consider pilots eligible for unit scheduling beyond a certain start state,
# which defaults to 'PMGR_ACTIVE'.
_BF_START = os.environ.get('RADICAL_PILOT_BACKFILLING_START', rps.PMGR_ACTIVE)
_BF_STOP  = os.environ.get('RADICAL_PILOT_BACKFILLING_STOP',  rps.PMGR_ACTIVE)

_BF_START_VAL = rps._pilot_state_value(_BF_START)
_BF_STOP_VAL  = rps._pilot_state_value(_BF_STOP)

# ==============================================================================
#
class Backfilling(UMGRSchedulingComponent):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        UMGRSchedulingComponent.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        self._wait_pool = dict()             # set of unscheduled units
        self._wait_lock = threading.RLock()  # look on the above set

        self._pids = list()
        self._idx  = 0


    # --------------------------------------------------------------------------
    #
    def add_pilots(self, pids):

      # self._log.debug('add pilots %s', pids)

        # pilots just got added.  If we did not have any pilot before, we might
        # have units in the wait queue waiting -- now is a good time to take
        # care of those!
        with self._wait_lock:

            # initialize custom data for the pilot
            for pid in pids:
                pilot = self._pilots[pid]['pilot']
                cores = pilot['description']['cores']
                hwm   = int(cores * _HWM/100)
                self._pilots[pid]['info'] = {
                        'cores' : cores,
                        'hwm'   : hwm,
                        'used'  : 0, 
                        'units' : list(), # list of assigned unit IDs
                        'done'  : list(), # list of executed unit IDs
                        }

            # now we can use the pilot
            self._pids += pids
            self._schedule_units()


    # --------------------------------------------------------------------------
    #
    def remove_pilots(self, pids):

      # self._log.debug('rem pilots %s', pids)

        with self._pilots_lock:

            for pid in pids:

                if not pid in self._pids:
                    raise ValueError('no such pilot %s' % pid)

                self._pids.remove(pid)
                # FIXME: cancel units


    # --------------------------------------------------------------------------
    #
    def update_pilots(self, pids):

      # self._log.debug('update pilots for %s', pids)

        # FIXME: if PMGR_ACTIVE: schedule
        # FIXME: if FINAL:  un/re-schedule
        action = False
        with self._pilots_lock:

            for pid in pids:

                state = self._pilots[pid]['state']

              # self._log.debug('=== update pilot: %s %s', pid, state)

                if  rps._pilot_state_value(state) < _BF_START_VAL:
                  # self._log.debug('=== early')
                    # not eligible, yet
                    continue

                if  rps._pilot_state_value(state) > _BF_STOP_VAL:
                  # self._log.debug('=== late')
                    # not eligible anymore
                    continue

                # this pilot is eligible.  Stop checking the others, and attempt
                # reschedule
                action = True
              # self._log.debug('break')
                break

      # self._log.debug('action: %s', action)

        if action:
          # self._log.debug('upd pilot  -> schedule')
            self._schedule_units()


    # --------------------------------------------------------------------------
    #
    def update_units(self, units):

      # self._log.debug('update  units: %s', [u['uid'] for u in units])

        reschedule = False

        with self._pilots_lock, self._wait_lock:

            for unit in units:
        
                uid   = unit['uid']
                state = unit['state']
                pid   = unit.get('pilot', '')

                self._log.debug('update  unit: %s [%s] [%s]',  uid, pid, state)

                if not pid:
                    # we are not interested in state updates for unscheduled
                    # units
                    self._log.debug('upd unit  %s no pilot', uid)
                    continue

                if not pid in self._pilots:
                    # we don't handle the pilot of this unit
                    self._log.debug('upd unit  %s not handled', uid)
                    continue

                info = self._pilots[pid]['info']

                if uid in info['done']:
                    # we don't need further state udates
                    self._log.debug('upd unit  %s in done', uid)
                    continue

                if  rps._unit_state_value(state) <= \
                    rps._unit_state_value(rps.AGENT_EXECUTING):
                    self._log.debug('upd unit  %s too early', uid)
                    continue

                if not uid in info['units']:
                    # this contradicts the unit's assignment
                    self._log.debug('upd unit  %s not in units', uid)
                    self._log.error('bf: unit %s on %s inconsistent', uid, pid)
                    raise RuntimeError('inconsistent scheduler state')

                # this unit is now considered done
                info['done'].append(uid)
                info['used'] -= unit['description']['cpu_processes'] \
                              * unit['description']['cpu_threads']
                reschedule = True
                self._log.debug('upd unit  %s -  schedule (used: %s)', uid, info['used'])

                if info['used'] < 0:
                    self._log.error('bf: pilot %s inconsistent', pid)
                    raise RuntimeError('inconsistent scheduler state')


        # if any pilot state was changed, consider new units for scheduling
        if reschedule:
            self._log.debug('upd units -> schedule')
            self._schedule_units()


    # --------------------------------------------------------------------------
    #
    def _work(self, units):

        with self._pilots_lock, self._wait_lock:

            for unit in units:

                uid = unit['uid']
                    
                # not yet scheduled - put in wait pool
                self._wait_pool[uid] = unit
                        
        self._schedule_units()


    # --------------------------------------------------------------------------
    #
    def _schedule_units(self):
        """
        We have a set of units which we can place over a set of pilots.  
        
        The overall objective is to keep pilots busy while load balancing across
        all pilots, even those which might yet to get added.  We achieve that
        via the following algorithm:

          - for each pilot which is being added, no matter the state:
            - assign sufficient units to the pilot that it can run 'n'
              generations of them, 'n' being a tunable parameter called
              'RADICAL_PILOT_BACKFILLING_HWM'.  

          - for each unit being completed (goes out of AGENT_EXECUTING state)
            - determine the pilot which executed it
            - backfill units from the wait queue until the backfilling HWM is
              reached again.

        The HWM is interpreted as percent of pilot size.  For example, a pilot
        of size 10 cores and a HWM of 200 can get units with a total of 20 cores
        assigned.  It can get assigned more than that, if the last unit
        assigned to it surpasses the HWM.  We will not schedule any unit larger
        than pilot size however.
        """

        with self._pilots_lock, self._wait_lock:

            # units to advance beyond scheduling
            to_advance = list()

            # check if we have pilots to schedule over
            if not self._pids:
                return

            # we ignore pilots which are not yet added, are not yet in
            # BF_START_STATE, and are beyond PMGR_ACTIVE state
            pids = list()
            for pid in self._pids:

                info  = self._pilots[pid]['info']
                state = self._pilots[pid]['state']
                role  = self._pilots[pid]['role']

                if role != ADDED:
                    continue

                if  rps._pilot_state_value(state) < _BF_START_VAL:
                    # not eligible, yet
                    continue

                if  rps._pilot_state_value(state) > _BF_STOP_VAL:
                    # not ligible anymore
                    continue

                if info['used'] >= info['hwm']:
                    # pilot is full
                    continue

                pids.append(pid)

            # check if we have any eligible pilots to schedule over
            if not pids:
                return

            # cycle over available pids and add units until we either ran
            # out of units to schedule, or out of pids to schedule over

          # self._log.debug('schedule %s units over %s pilots',
          #         len(self._wait_pool), len(pids))

            scheduled   = list()   # units we want to advance
            unscheduled = dict()   # this will be the new wait pool
            for uid, unit in self._wait_pool.iteritems():

                if not pids:
                    # no more useful pilots -- move remaining units into
                    # unscheduled pool
                    self._log.debug(' =!= sch unit  %s', uid)
                    unscheduled[uid] = unit
                    continue

                cores   = unit['description']['cpu_processes'] \
                        * unit['description']['cpu_threads']
                success = False
                for pid in pids:

                    info = self._pilots[pid]['info']

                    if info['used'] <= info['hwm']:

                      # self._log.debug('sch unit  %s -> %s', uid, pid)
                        self._log.info('schedule %s -> %s', uid, pid)

                        pilot = self._pilots[pid]['pilot']
                        info['units'].append(unit['uid'])
                        info['used']   += cores

                        self._assign_pilot(unit, pilot)
                        scheduled.append(unit)
                        success = True

                        # this pilot might now be full.  If so, remove it from
                        # list of eligible pids
                        if info['used'] >= info['hwm']:
                            pids.remove(pid)

                        break  # stop looking through pilot list

                if not success:
                    # we did not find a useable pilot for this unit -- keep it
                    self._log.debug(' ==! sch unit  %s', uid)
                    unscheduled[uid] = unit


          # self._log.debug('retain   %s units and  %s pilots',
          #         len(unscheduled), len(pids))

            # all unscheduled units *are* the new wait pool
            self._log.debug(' 1 > waits: %s', self._wait_pool.keys())
            self._wait_pool = unscheduled
            self._log.debug(' 2 > waits: %s', self._wait_pool.keys())

        # advance scheduled units
        if scheduled:
            self.advance(scheduled, rps.UMGR_STAGING_INPUT_PENDING, 
                         publish=True, push=True)


        self._log.debug('\nafter schedule:')
        self._log.debug('waits:    %s', self._wait_pool.keys())
      # for pid in self._pilots:
      #     print 'pilot %s' % pid
      #     pprint.pprint(self._pilots[pid]['info'])
      # self._log.debug()
        

# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os
import threading
import subprocess    as sp

import radical.utils as ru

from ... import states    as rps
from ... import constants as rpc

from .base import AgentExecutingComponent


# ==============================================================================
#
class ShellFS(AgentExecutingComponent):


    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        AgentExecutingComponent.__init__ (self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize_child(self):

        from .... import pilot as rp

        self._pwd = os.getcwd() 
        self._tmp = self._pwd   # keep temporary files in $PWD for now (slow)

        self.register_input(rps.AGENT_EXECUTING_PENDING,
                            rpc.AGENT_EXECUTING_QUEUE, self.work)

        self.register_output(rps.AGENT_STAGING_OUTPUT_PENDING,
                             rpc.AGENT_STAGING_OUTPUT_QUEUE)

        self.register_publisher (rpc.AGENT_UNSCHEDULE_PUBSUB)
        self.register_subscriber(rpc.CONTROL_PUBSUB, self.command_cb)

        # Mimic what virtualenv's "deactivate" would do
        self._deactivate = "\n# deactivate pilot virtualenv\n"

        old_path  = os.environ.get('_OLD_VIRTUAL_PATH',       None)
        old_ppath = os.environ.get('_OLD_VIRTUAL_PYTHONPATH', None)
        old_home  = os.environ.get('_OLD_VIRTUAL_PYTHONHOME', None)
        old_ps1   = os.environ.get('_OLD_VIRTUAL_PS1',        None)

        if old_ppath: self._deactivate += 'export PATH="%s"\n'        % old_ppath
        if old_path : self._deactivate += 'export PYTHONPATH="%s"\n'  % old_path
        if old_home : self._deactivate += 'export PYTHON_HOME="%s"\n' % old_home
        if old_ps1  : self._deactivate += 'export PS1="%s"\n'         % old_ps1

        self._deactivate += 'unset VIRTUAL_ENV\n\n'

        # FIXME: we should not alter the environment of the running agent, but
        #        only make sure that the CU finds a pristine env.  That also
        #        holds for the unsetting below -- AM
        if old_path : os.environ['PATH']        = old_path
        if old_ppath: os.environ['PYTHONPATH']  = old_ppath
        if old_home : os.environ['PYTHON_HOME'] = old_home
        if old_ps1  : os.environ['PS1']         = old_ps1

        if 'VIRTUAL_ENV' in os.environ :
            del(os.environ['VIRTUAL_ENV'])

        self._task_launcher = None
        self._mpi_launcher  = None

        try:
            self._task_launcher = rp.agent.LM.create(
                    name    = self._cfg['task_launch_method'],
                    cfg     = self._cfg,
                    session = self._session)
        except:
            self._log.warn('no task launcher found')

        try:
            self._mpi_launcher = rp.agent.LM.create(
                    name    = self._cfg['mpi_launch_method'],
                    cfg     = self._cfg,
                    session = self._session)
        except:
            self._log.warn('no mpi launcher found')

        # TODO: test that this actually works
        # Remove the configured set of environment variables from the
        # environment that we pass to Popen.
        for e in os.environ.keys():
            env_removables = list()
            if self._mpi_launcher : env_removables += self._mpi_launcher.env_removables
            if self._task_launcher: env_removables += self._task_launcher.env_removables
            for r in  env_removables:
                if e.startswith(r):
                    os.environ.pop(e, None)

        # if we need to transplant any original env into the CU, we dig the
        # respective keys from the dump made by bootstrap_0.sh
        self._env_cu_export = dict()
        if self._cfg.get('export_to_cu'):
            with open('env.orig', 'r') as f:
                for line in f.readlines():
                    if '=' in line:
                        k,v = line.split('=', 1)
                        key = k.strip()
                        val = v.strip()
                        if key in self._cfg['export_to_cu']:
                            self._env_cu_export[key] = val

        # the registry keeps track of units to watch
        self._registry      = dict()
        self._registry_lock = threading.RLock()

        self._to_cancel  = list()
        self._cancel_lock    = threading.RLock()

        self._cached_events = list()  # keep monitoring events for pid's which
                                      # are not yet known

        self.gtod = "%s/gtod" % self._pwd

        # create line buffered fifo's to communicate with the shell executor
        self._fifo_cmd_name = "%s/%s.cmd.pipe" % (self._tmp, self._uid)
        self._fifo_inf_name = "%s/%s.inf.pipe" % (self._tmp, self._uid)

        os.mkfifo(self._fifo_cmd_name)
        os.mkfifo(self._fifo_inf_name)

        self._fifo_cmd = open(self._fifo_cmd_name, 'w+', 1)
        self._fifo_inf = open(self._fifo_inf_name, 'r+', 1)

        # run thread to watch then info fifo
        self._terminate = threading.Event()
        self._watcher   = threading.Thread(target=self._watch, name="Watcher")
        self._watcher.daemon = True
        self._watcher.start ()

        # start the shell executor
        sh_exe   = "%s/shell_spawner_fs.sh" % os.path.dirname(__file__)
        sh_cmd   = "%s %s %s %s" % (sh_exe, self._pwd, self._tmp, self.uid)
                                 # script   base       work       sid
        self._log.debug('start shell executor [%s]', sh_cmd)
        self._sh = sp.Popen(sh_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)


    # --------------------------------------------------------------------------
    #
    def finalize_child(self):

        self._terminate.set()

        try:
            self._sh.kill()
        except:
            pass

        try:
            os.unlink(self._fifo_cmd_name)
            os.unlink(self._fifo_inf_name)
            self._fifo_cmd.close()
            self._fifo_inf.close()
        except:
            pass


    # --------------------------------------------------------------------------
    #
    def command_cb(self, topic, msg):

        cmd = msg['cmd']
        arg = msg['arg']

        if cmd == 'cancel_units':

            self._log.info("cancel_units command (%s)" % arg)
            uids = arg['uids']

            if not isinstance(uids, list):
                uids = [uids]

            with self._cancel_lock:
                for uid in uids:
                    self._fifo_cmd.write('KILL %s\n' % uid)
                    self._to_cancel.append(uid)

            self._fifo_cmd.flush()

            # The state advance will be managed by the watcher, which will pick
            # up the cancel notification.
            # FIXME: We could optimize a little by publishing the unschedule
            #        right here...

        return True


    # --------------------------------------------------------------------------
    #
    def work(self, units):

        if not isinstance(units, list):
            units = [units]

        self.advance(units, rps.AGENT_EXECUTING, publish=True, push=False)

        for unit in units:
            self._handle_unit(unit)

        # fail on dead watcher
        if self._terminate.is_set():
            raise RuntimeError('watcher died')


    # --------------------------------------------------------------------------
    #
    def _handle_unit(self, cu):

        # check that we don't start any units which need cancelling
        if cu['uid'] in self._to_cancel:

            with self._cancel_lock:
                self._to_cancel.remove(cu['uid'])

            self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, cu)
            self.advance(cu, rps.CANCELED, publish=True, push=False)
            return True

        # launch the new unit
        try:
            mpi = cu['description'].get('mpi', False) 
            if mpi: launcher = self._mpi_launcher
            else  : launcher = self._task_launcher

            if not launcher:
                raise RuntimeError("no launcher (mpi=%s)" % mpi)

            self._log.debug("Launching with %s (%s).", launcher.name,
                            launcher.launch_command)

            assert(cu['slots'])

            self.spawn(launcher=launcher, cu=cu)


        except Exception as e:
            # append the startup error to the units stderr.  This is
            # not completely correct (as this text is not produced
            # by the unit), but it seems the most intuitive way to
            # communicate that error to the application/user.
            self._log.exception("error running CU: %s", e)

            # Free the Slots, Flee the Flots, Ree the Frots!
            if cu.get('slots'):
                self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, cu)

            self.advance(cu, rps.FAILED, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def _cu_to_cmd (self, cu, launcher) :

        # ----------------------------------------------------------------------
        def quote_args (args) :

            ret = list()
            for arg in args :

                if not arg:
                    continue

                # if string is between outer single quotes,
                #    pass it as is.
                # if string is between outer double quotes,
                #    pass it as is.
                # otherwise (if string is not quoted)
                #    escape all double quotes

                if  arg[0] == arg[-1]  == "'" :
                    ret.append (arg)
                elif arg[0] == arg[-1] == '"' :
                    ret.append (arg)
                else :
                    arg = arg.replace ('"', '\\"')
                    ret.append ('"%s"' % arg)

            return  ret

        # ----------------------------------------------------------------------

        args  = ""
        env   = self._deactivate
        cwd   = ""
        pre   = ""
        post  = ""
        io    = ""
        cmd   = ""
        descr = cu['description']

        sandbox  = '%s/%s' % (self._pwd, cu['uid'])

        env  += "# CU environment\n"
        env  += "export RP_SESSION_ID=%s\n"     % self._cfg['session_id']
        env  += "export RP_PILOT_ID=%s\n"       % self._cfg['pilot_id']
        env  += "export RP_AGENT_ID=%s\n"       % self._cfg['agent_name']
        env  += "export RP_SPAWNER_ID=%s\n"     % self.uid
        env  += "export RP_UNIT_ID=%s\n"        % cu['uid']
        env  += 'export RP_GTOD="%s"\n'         % self.gtod
        if 'RADICAL_PILOT_PROFILE' in os.environ or \
           'RADICAL_PROFILE'       in os.environ :
            env += 'export RP_PROF="%s/%s.prof"\n' % (sandbox, cu['uid'])
        env  += '''
prof(){
    test -z "$RP_PROF" && return
    event=$1
    now=$($RP_GTOD)
    echo "$now,$event,unit_script,MainThread,$RP_UNIT_ID,AGENT_EXECUTING," >> $RP_PROF
}
'''

        # also add any env vars requested for export by the resource config
        for k,v in self._env_cu_export.iteritems():
            env += "export %s=%s\n" % (k,v)

        # also add any env vars requested in hte unit description
        if descr['environment']:
            for e in descr['environment'] :
                env += "export %s=%s\n"  %  (e, descr['environment'][e])
        env  += "\n"

        cwd  += "# CU sandbox\n"
        cwd  += "mkdir -p %s\n" % sandbox
        cwd  += "cd       %s\n" % sandbox
        cwd  += 'prof cu_cd_done\n'
        cwd  += "\n"

        if  descr['pre_exec'] :
            fail  = ' (echo "pre_exec failed"; false) || exit'
            pre  += "\n# CU pre-exec\n"
            pre  += 'prof cu_pre_start\n'
            for elem in descr['pre_exec']:
                pre += "%s || %s\n" % (elem, fail)
            pre  += "\n"
            pre  += 'prof cu_pre_stop\n'
            pre  += "\n"

        if  descr['post_exec'] :
            fail  = ' (echo "post_exec failed"; false) || exit'
            post += "\n# CU post-exec\n"
            post += 'prof cu_post_start\n'
            for elem in descr['post_exec']:
                post += "%s || %s\n" % (elem, fail)
            post += 'prof cu_post_stop\n'
            post += "\n"

        if  descr['arguments']  :
            args  = ' ' .join (quote_args (descr['arguments']))

        stdout_file = descr.get('stdout') or 'STDOUT'
        stderr_file = descr.get('stderr') or 'STDERR'

        cu['stdout_file'] = os.path.join(sandbox, stdout_file)
        cu['stderr_file'] = os.path.join(sandbox, stderr_file)

        io  += "1>%s " % cu['stdout_file']
        io  += "2>%s " % cu['stderr_file']

        cmd, hop_cmd  = launcher.construct_command(cu, '/usr/bin/env RP_SPAWNER_HOP=TRUE "$0"')

        script  = '\n%s\n' % env
        script += 'prof cu_start\n'

        if hop_cmd :
            # the script will itself contain a remote callout which calls again
            # the script for the invokation of the real workload (cmd) -- we
            # thus introduce a guard for the first execution.  The hop_cmd MUST
            # set RP_SPAWNER_HOP to some value for the startup to work

            script += "\n# ------------------------------------------------------\n"
            script += '# perform one hop for the actual command launch\n'
            script += 'if test -z "$RP_SPAWNER_HOP"\n'
            script += 'then\n'
            script += '    %s\n' % hop_cmd
            script += '    exit\n'
            script += 'fi\n\n'

        script += "\n# ------------------------------------------------------\n"
        script += "%s"        %  cwd
        script += "%s"        %  pre
        script += "\n# CU execution\n"
        script += 'prof cu_exec_start\n'
        script += "%s %s\n\n" % (cmd, io)
        script += "RETVAL=$?\n"
        script += 'prof cu_exec_stop\n'
        script += "%s"        %  post
        script += "# notify the agent\n"
        script += "echo \"FINAL %s $RETVAL\" > %s\n" % (cu['uid'],
                                                        self._fifo_inf_name)
        script += "exit $RETVAL\n"
        script += "# ------------------------------------------------------\n\n"

      # self._log.debug ("execution script:\n%s\n", script)

        return script


    # --------------------------------------------------------------------------
    #
    def spawn(self, launcher, cu):

        uid     = cu['uid']
        sandbox = '%s/%s' % (self._pwd, uid)

        try:
            os.makedirs(sandbox)
        except:
            pass

        # prep stdout/err so that we can append w/o checking for None
        cu['stdout'] = ''
        cu['stderr'] = ''

        # we got an allocation: go off and launch the process.  we get
        # a multiline command, so use the wrapper's BULK/LRUN mode.
        cmd = self._cu_to_cmd (cu, launcher)
        with open("%s/%s.sh" % (sandbox, uid), 'w+') as fout:
            fout.write(cmd)

        self._prof.prof('exec_start', uid=uid)
        self._fifo_cmd.write('EXEC %s\n' % uid)
        self._fifo_cmd.flush()

        with self._registry_lock :
            self._registry[uid] = cu

        # FIXME: HERE

    # --------------------------------------------------------------------------
    #
    def _readline(self, timeout=None):

        # NOTE: this will only work for line-buffered I/O

        import select

        r, _, _ = select.select([self._fifo_inf], [], [], timeout)

        if r:
            return self._fifo_inf.readline()


    # --------------------------------------------------------------------------
    #
    def _watch (self) :

        TIMEOUT = 1.0   # check for stop signal now and then

        while not self._terminate.is_set():

            try:
                line     = self._readline(TIMEOUT)

                if not line:
                    continue

                cmd, msg = line.split(' ', 1)

                if cmd == 'FINAL':
                    self._handle_event (msg)

                elif line == 'EXIT' or line == "Killed" :
                    self._log.error ("monitoring channel failed (%s)", msg)
                    self._terminate.set()
                    return

            except Exception as e:
                self._log.exception("Exception in job monitoring thread: %s", e)
                self._terminate.set()


    # --------------------------------------------------------------------------
    #
    def _handle_event (self, msg):

        if ' ' in msg:
            uid, ret = msg.split(' ', 1)
        else:
            uid = msg.strip()
            ret = None

        self._prof.prof('exec_stop', uid=uid)

        with self._registry_lock:
            cu = self._registry[uid]
            del(self._registry[uid])

        # free unit slots.
        self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, cu)

        if ret is None:
            cu['exit_code'] = None
        else:
            cu['exit_code'] = int(ret)

        if   cu['exit_code'] == 0   : rp_state = rps.DONE
        elif cu['exit_code'] is None: rp_state = rps.CANCELED
        else                        : rp_state = rps.FAILED

        if rp_state in [rps.FAILED, rps.CANCELED] :
            # The unit failed - fail after staging output
            cu['target_state'] = rps.FAILED

        else:
            # The unit finished cleanly, see if we need to deal with
            # output data.  We always move to stageout, even if there are no
            # directives -- at the very least, we'll upload stdout/stderr
            cu['target_state'] = rps.DONE

        self.advance(cu, rps.AGENT_STAGING_OUTPUT_PENDING, publish=True, push=True)


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'

import os
import sys

import radical.pilot as rp
import radical.utils as ru

rpu = rp.utils


#------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) <= 1:
        print "\n\tusage: %s <session_id>\n"
        sys.exit(1)

    sid = sys.argv[1]


    profiles = rpu.fetch_profiles(sid=sid, skip_existing=True)
    for p in profiles:
        print p

    profs = rpu.read_profiles(profiles)

    for p in profs:
        print type(p)

    prof = rpu.combine_profiles(profs)

    print len(prof)
    for entry in prof:
        print entry




#-------------------------------------------------------------------------------

